package org.hardknockjava.lesson15.record;

/*******************************************************************
 * BaseComparator compares BaseRecord instances
 * according to the keys specified in the BaseRecord
 * subclass. Note that BaseComparator uses reflection
 * to determine what the keys and field values are.
 * It handles any subclass of BaseRecord without
 * the need to know the names or types of its
 * fields.
 */

import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public class BaseComparator implements Comparator<BaseRecord> {
    // controlMap contains SortControl lists for each class that's
    // been sorted.
    private Map<Class<? extends BaseRecord>, List<SortControl>> controlMap = new HashMap<Class<? extends BaseRecord>, List<SortControl>>();

    @SuppressWarnings("unchecked")
    @Override
    public int compare(BaseRecord o1, BaseRecord o2) {
        // Get the list of SortControls for the class being compared.
        // If it doesn't exist, create it and save it in the map.
        List<SortControl> controlList = controlMap.get(o1.getClass());
        if (controlList == null) {
            controlList = createControlList(o1.getClass());
            controlMap.put(o1.getClass(), controlList);
        }

        // This is the return value.
        int comparison = 0;

        // Sort on each key field in turn until comparison
        // is nonzero or the fields are exhausted.
        for (SortControl sortControl : controlList) {
            Field sortField = sortControl.getSortField();

            // Save the current accessibility of the sort field
            // and ensure it's accessible.
            boolean accessibility = sortField.canAccess(o1);
            sortField.setAccessible(true);

            Object field1 = null;
            Object field2 = null;

            try {
                // TODO
                // Get the values of the source field.
                // field1 = ?
                // field2 = ?

                // Set comparison to the compare values of the field.
                // If fields1 is null then if field2 is not, set comparison
                // to 1; if field2 is also null, leave it 0.
                if (field1 == null) {
                    if (field2 != null) {
                        comparison = 1;
                    }
                } else {
                    // If the fields are comparable, set comparison to the
                    // value of compareTo().
                    if (field2 == null) {
                        comparison = -1;
                    } else {
                        if (field1 instanceof Comparable<?>) {
                            comparison = ((Comparable<Object>) field1).compareTo((Comparable<Object>) field2);
                        }
                    }

                    // If the field is being sorted in descending order, flip
                    // the value of comparison.
                    if (sortControl.getSequence().equals(SequenceType.DESCENDING)) {
                        comparison = -1 * comparison;
                    }

                    // If comparison is nonzero (we've found unequal field values),
                    // break out of the loop.
                    if (comparison != 0) {
                        break;
                    }
                }
            } catch (Exception e) {
            } finally {
                // Whatever happens, restore the original accessibility value.
                sortField.setAccessible(accessibility);
            }
        }

        return comparison;
    }

    // When a BaseRecord subclass is sorted for the first time,
    // build a list of SortControls from the @Key fields in
    // the class provided and its superclasses.
    @SuppressWarnings("unchecked")
    private List<SortControl> createControlList(Class<? extends BaseRecord> clazz) {
        // Find the fields with a @Key annotation and save them
        // in a List.
        List<SortControl> controlList = new ArrayList<SortControl>();

        while (clazz != null) {
            // TODO
            // Build elements of controlList from the fields
            // with the @Key annotation. Remember to check for
            // sort order and sequence type.
            
            // Repeat for the superclass.
            clazz = (Class<? extends BaseRecord>) clazz.getSuperclass();
        }

        // Sort the SortControls by the order field, which is their natural order.
        controlList.sort(null);

        return controlList;
    }
}
